@@ -1,7 +1,10 @@
 use crate::built_in_types::BuiltInType;
-use crate::parse::HostLang;
+use crate::parse::{HostLang, TypeDeclarations};
 use crate::parsed_extern_fn::ParsedExternFn;
-use crate::{BridgedType, OpaqueForeignType, SwiftBridgeModule, SWIFT_BRIDGE_PREFIX};
+use crate::{
+    BridgedType, OpaqueForeignType, SharedType, StructSwiftRepr, SwiftBridgeModule,
+    SWIFT_BRIDGE_PREFIX,
+};
 use quote::ToTokens;
 use std::collections::HashMap;
 use std::ops::Deref;
@@ -37,22 +40,33 @@ impl SwiftBridgeModule {
             }
 
             let func_definition = match function.host_lang {
-                HostLang::Rust => gen_func_swift_calls_rust(function),
-                HostLang::Swift => gen_function_exposes_swift_to_rust(function),
+                HostLang::Rust => gen_func_swift_calls_rust(function, &self.types),
+                HostLang::Swift => gen_function_exposes_swift_to_rust(function, &self.types),
             };
 
             swift += &func_definition;
             swift += "\n";
         }
 
-        for ty in &self.types {
+        for ty in self.types.types() {
             let generated = match ty {
-                BridgedType::Shared(_) => {
-                    //
-                    todo!()
+                BridgedType::Shared(SharedType::Struct(shared_struct)) => {
+                    match shared_struct.swift_repr {
+                        StructSwiftRepr::Class => {
+                            todo!()
+                        }
+                        StructSwiftRepr::Structure => {
+                            // No need to generate any code. Swift will automatically generate a
+                            //  struct from our C header typedef that we generate for this struct.
+
+                            continue;
+                        }
+                    }
                 }
                 BridgedType::Opaque(ty) => match ty.host_lang {
-                    HostLang::Rust => generate_swift_class(ty, &associated_funcs_and_methods),
+                    HostLang::Rust => {
+                        generate_swift_class(ty, &associated_funcs_and_methods, &self.types)
+                    }
                     HostLang::Swift => generate_drop_swift_instance_reference_count(ty),
                 },
             };
@@ -68,6 +82,7 @@ impl SwiftBridgeModule {
 fn generate_swift_class(
     ty: &OpaqueForeignType,
     associated_funcs_and_methods: &HashMap<String, Vec<&ParsedExternFn>>,
+    types: &TypeDeclarations,
 ) -> String {
     let type_name = ty.ident.to_string();
 
@@ -82,7 +97,7 @@ fn generate_swift_class(
         for type_method in methods {
             // TODO: Normalize with freestanding func codegen above
 
-            let func_definition = gen_func_swift_calls_rust(type_method);
+            let func_definition = gen_func_swift_calls_rust(type_method, types);
 
             if type_method.is_initializer {
                 initializers.push(func_definition);
@@ -159,10 +174,10 @@ func {fn_name} (ptr: UnsafeMutableRawPointer) {{
     )
 }
 
-fn gen_func_swift_calls_rust(function: &ParsedExternFn) -> String {
+fn gen_func_swift_calls_rust(function: &ParsedExternFn, types: &TypeDeclarations) -> String {
     let fn_name = function.sig.ident.to_string();
-    let params = function.to_swift_param_names_and_types(false);
-    let call_args = function.to_swift_call_args(true, false);
+    let params = function.to_swift_param_names_and_types(false, types);
+    let call_args = function.to_swift_call_args(true, false, types);
     let call_fn = format!("{}({})", fn_name, call_args);
 
     let type_name_segment = if let Some(ty) = function.associated_type.as_ref() {
@@ -190,7 +205,7 @@ fn gen_func_swift_calls_rust(function: &ParsedExternFn) -> String {
     let maybe_return = if function.is_initializer {
         "".to_string()
     } else {
-        function.to_swift_return_type(false)
+        function.to_swift_return_type(false, types)
     };
 
     let swift_class_func_name = if function.is_initializer {
@@ -225,13 +240,24 @@ fn gen_func_swift_calls_rust(function: &ParsedExternFn) -> String {
                     unreachable!()
                 }
                 ReturnType::Type(_, ty) => {
-                    let (is_owned, ty) = match ty.deref() {
-                        Type::Reference(reference) => ("false", &reference.elem),
-                        _ => ("true", ty),
+                    let ty_name = match ty.deref() {
+                        Type::Reference(reference) => reference.elem.to_token_stream().to_string(),
+                        Type::Path(path) => path.path.segments.to_token_stream().to_string(),
+                        _ => todo!(),
                     };
 
-                    let ty = ty.to_token_stream().to_string();
-                    format!("{}(ptr: {}, isOwned: {})", ty, call_rust, is_owned)
+                    match types.get(&ty_name).unwrap() {
+                        BridgedType::Shared(_) => call_rust,
+                        BridgedType::Opaque(_) => {
+                            let (is_owned, ty) = match ty.deref() {
+                                Type::Reference(reference) => ("false", &reference.elem),
+                                _ => ("true", ty),
+                            };
+
+                            let ty = ty.to_token_stream().to_string();
+                            format!("{}(ptr: {}, isOwned: {})", ty, call_rust, is_owned)
+                        }
+                    }
                 }
             }
         }
@@ -252,15 +278,15 @@ fn gen_func_swift_calls_rust(function: &ParsedExternFn) -> String {
     func_definition
 }
 
-fn gen_function_exposes_swift_to_rust(func: &ParsedExternFn) -> String {
+fn gen_function_exposes_swift_to_rust(func: &ParsedExternFn, types: &TypeDeclarations) -> String {
     let link_name = func.link_name();
     let prefixed_fn_name = func.prefixed_fn_name();
     let fn_name = &func.sig.ident;
 
-    let params = func.to_swift_param_names_and_types(true);
-    let ret = func.to_swift_return_type(true);
+    let params = func.to_swift_param_names_and_types(true, types);
+    let ret = func.to_swift_return_type(true, types);
 
-    let args = func.to_swift_call_args(false, true);
+    let args = func.to_swift_call_args(false, true, types);
     let mut call_fn = format!("{}({})", fn_name, args);
 
     if let Some(associated_type) = func.associated_type.as_ref() {
@@ -994,7 +1020,7 @@ func get_owned_foo() -> Foo {
 
     /// Verify that we can return a reference to a declared type.
     #[test]
-    fn extern_rust_return_reference_to_declared_declared_type() {
+    fn extern_rust_return_reference_to_declared_type() {
         let tokens = quote! {
             mod foo {
                 extern "Rust" {
@@ -1011,6 +1037,80 @@ func get_owned_foo() -> Foo {
 func get_ref_foo() -> Foo {
     Foo(ptr: __swift_bridge__$get_ref_foo(), isOwned: false)
 }
+"#;
+
+        assert_generated_contains_expected(&generated, &expected);
+    }
+
+    /// Verify that we can return a shared struct type.
+    #[test]
+    fn extern_rust_return_shared_struct() {
+        let tokens = quote! {
+            mod ffi {
+                struct Foo;
+
+                extern "Rust" {
+                    fn get_foo () -> Foo;
+                }
+            }
+        };
+        let module: SwiftBridgeModule = parse_quote!(#tokens);
+        let generated = module.generate_swift();
+
+        let expected = r#"
+func get_foo() -> Foo {
+    __swift_bridge__$get_foo()
+}
+"#;
+
+        assert_generated_contains_expected(&generated, &expected);
+    }
+
+    /// Verify that we can take a shared struct as an argument.
+    #[test]
+    fn extern_rust_shared_struct_arg() {
+        let tokens = quote! {
+            mod ffi {
+                struct Foo;
+
+                extern "Rust" {
+                    fn some_function (arg: Foo);
+                }
+            }
+        };
+        let module: SwiftBridgeModule = parse_quote!(#tokens);
+        let generated = module.generate_swift();
+
+        let expected = r#"
+func some_function(_ arg: Foo) {
+    __swift_bridge__$some_function(arg)
+}
+"#;
+
+        assert_generated_contains_expected(&generated, &expected);
+    }
+
+    /// Verify that we rename shared struct arguments and return values if there is a swift_name
+    /// attribute.
+    #[test]
+    fn extern_rust_fn_uses_swift_name_for_shared_struct_attrib() {
+        let tokens = quote! {
+            mod ffi {
+                #[swift_bridge(swift_name = "Renamed")]
+                struct Foo;
+
+                extern "Rust" {
+                    fn some_function (arg: Foo) -> Foo;
+                }
+            }
+        };
+        let module: SwiftBridgeModule = parse_quote!(#tokens);
+        let generated = module.generate_swift();
+
+        let expected = r#"
+func some_function(_ arg: Renamed) -> Renamed {
+    __swift_bridge__$some_function(arg)
+}
 "#;
 
         assert_generated_contains_expected(&generated, &expected);
