@@ -1,6 +1,6 @@
 use crate::built_in_types::BuiltInType;
-use crate::parse::HostLang;
-use crate::{pat_type_pat_is_self, BridgedType, SWIFT_BRIDGE_PREFIX};
+use crate::parse::{HostLang, TypeDeclarations};
+use crate::{pat_type_pat_is_self, BridgedType, SharedType, SWIFT_BRIDGE_PREFIX};
 use proc_macro2::{Ident, TokenStream};
 use quote::{quote, quote_spanned, ToTokens};
 use std::ops::Deref;
@@ -53,7 +53,11 @@ impl ParsedExternFn {
         }
     }
 
-    pub(crate) fn rust_return_type(&self, swift_bridge_path: &Path) -> TokenStream {
+    pub(crate) fn rust_return_type(
+        &self,
+        swift_bridge_path: &Path,
+        types: &TypeDeclarations,
+    ) -> TokenStream {
         let sig = &self.func.sig;
 
         let ret = match &sig.output {
@@ -73,15 +77,24 @@ impl ParsedExternFn {
                             _ => (false, ty.deref()),
                         };
 
-                        let ptr = if is_const_ptr {
-                            quote! { *const }
-                        } else {
-                            quote! { *mut }
-                        };
+                        let ty_string = ty.to_token_stream().to_string();
+                        match types.get(&ty_string).unwrap() {
+                            BridgedType::Shared(SharedType::Struct(shared_struct)) => {
+                                let name = &shared_struct.name;
+                                quote! { #arrow #name }
+                            }
+                            BridgedType::Opaque(_) => {
+                                let ptr = if is_const_ptr {
+                                    quote! { *const }
+                                } else {
+                                    quote! { *mut }
+                                };
 
-                        quote_spanned! {ty.span()=> -> #ptr super::#ty }
+                                quote_spanned! {ty.span()=> -> #ptr super::#ty }
+                            }
+                        }
                     } else {
-                        quote_spanned! {ty.span()=> -> *mut std::ffi::c_void }
+                        quote_spanned! {ty.span()=> #arrow *mut std::ffi::c_void }
                     }
                 }
             }
@@ -123,7 +136,11 @@ impl ParsedExternFn {
     // fn foo (&self, arg1: u8, arg2: u32, &SomeType)
     //  becomes..
     // self.0, arg1, arg2, & unsafe { Box::from_raw(bar }
-    pub fn to_call_rust_args(&self, swift_bridge_path: &Path) -> TokenStream {
+    pub fn to_call_rust_args(
+        &self,
+        swift_bridge_path: &Path,
+        types: &TypeDeclarations,
+    ) -> TokenStream {
         let mut args = vec![];
         let inputs = &self.func.sig.inputs;
         for arg in inputs {
@@ -160,13 +177,34 @@ impl ParsedExternFn {
                             );
                         };
                     } else {
-                        let (maybe_ref, maybe_mut) = match pat_ty.ty.deref() {
-                            Type::Reference(ty_ref) => (Some(ty_ref.and_token), ty_ref.mutability),
-                            _ => (None, None),
+                        let ty_string = match pat_ty.ty.deref() {
+                            Type::Reference(reference) => {
+                                reference.elem.to_token_stream().to_string()
+                            }
+                            Type::Path(ty_path) => {
+                                ty_path.path.segments.to_token_stream().to_string()
+                            }
+                            _ => todo!(),
                         };
+                        match types.get(&ty_string).unwrap() {
+                            BridgedType::Shared(SharedType::Struct(_)) => {}
+                            BridgedType::Opaque(_) => {
+                                let (maybe_ref, maybe_mut) = match pat_ty.ty.deref() {
+                                    Type::Reference(ty_ref) => {
+                                        (Some(ty_ref.and_token), ty_ref.mutability)
+                                    }
+                                    _ => (None, None),
+                                };
 
-                        let dereferenced = quote! { unsafe { #maybe_ref #maybe_mut * #arg } };
-                        arg = dereferenced;
+                                let dereferenced = if maybe_ref.is_some() {
+                                    quote! { unsafe { #maybe_ref #maybe_mut * #arg } }
+                                } else {
+                                    quote! { unsafe { *Box::from_raw(#arg) } }
+                                };
+
+                                arg = dereferenced;
+                            }
+                        };
                     }
 
                     args.push(arg);
@@ -182,7 +220,7 @@ impl ParsedExternFn {
     // fn foo (&self, arg1: u8, arg2: u32)
     //  becomes..
     // void* self, uint8_t u8, uint32_t arg2
-    pub fn to_c_header_params(&self) -> String {
+    pub fn to_c_header_params(&self, types: &TypeDeclarations) -> String {
         let mut params = vec![];
         let inputs = &self.func.sig.inputs;
         for arg in inputs {
@@ -195,9 +233,16 @@ impl ParsedExternFn {
                         params.push("void* self".to_string());
                     } else {
                         let ty = if let Some(built_in) = BuiltInType::new_with_type(&pat_ty.ty) {
-                            built_in.to_c().to_string()
+                            built_in.to_c()
                         } else {
-                            pat.to_token_stream().to_string()
+                            let ty = &pat_ty.ty;
+                            let ty_string = ty.to_token_stream().to_string();
+                            match types.get(&ty_string).unwrap() {
+                                BridgedType::Shared(SharedType::Struct(shared_struct)) => {
+                                    format!("struct {}", shared_struct.swift_name_string())
+                                }
+                                BridgedType::Opaque(_) => "void*".to_string(),
+                            }
                         };
 
                         let arg_name = pat.to_token_stream().to_string();
@@ -214,14 +259,25 @@ impl ParsedExternFn {
         }
     }
 
-    pub fn to_c_header_return(&self) -> String {
+    pub fn to_c_header_return(&self, types: &TypeDeclarations) -> String {
         match &self.func.sig.output {
             ReturnType::Default => "void".to_string(),
             ReturnType::Type(_, ty) => {
                 if let Some(ty) = BuiltInType::new_with_type(&ty) {
                     ty.to_c()
                 } else {
-                    "void*".to_string()
+                    let ty_string = match ty.deref() {
+                        Type::Reference(reference) => reference.elem.to_token_stream().to_string(),
+                        Type::Path(path) => path.path.to_token_stream().to_string(),
+                        _ => todo!(),
+                    };
+
+                    match types.get(&ty_string).unwrap() {
+                        BridgedType::Shared(SharedType::Struct(shared_struct)) => {
+                            format!("struct {}", shared_struct.swift_name_string())
+                        }
+                        BridgedType::Opaque(_) => "void*".to_string(),
+                    }
                 }
             }
         }
@@ -326,7 +382,7 @@ mod tests {
     use super::*;
     use crate::errors::{ParseError, ParseErrors};
     use crate::parse::SwiftBridgeModuleAndErrors;
-    use crate::test_utils::assert_tokens_contain;
+    use crate::test_utils::{assert_tokens_contain, assert_tokens_eq};
     use crate::SwiftBridgeModule;
 
     /// Verify that when generating rust call args we do not include the receiver.
@@ -352,7 +408,7 @@ mod tests {
 
         for method in methods {
             let rust_call_args =
-                &method.to_call_rust_args(&syn::parse2(quote! { swift_bridge }).unwrap());
+                &method.to_call_rust_args(&module.swift_bridge_path, &module.types);
             assert_eq!(
                 rust_call_args.to_string(),
                 "",
@@ -362,6 +418,25 @@ mod tests {
         }
     }
 
+    /// Verify that when we get an owned opaque type as an argument we unbox it.
+    #[test]
+    fn unboxes_owned_opaque_type() {
+        let tokens = quote! {
+            #[swift_bridge::bridge]
+            mod ffi {
+                extern "Rust" {
+                    type Foo;
+                    fn some_function (arg: Foo);
+                }
+            }
+        };
+        let module = parse_ok(tokens);
+        assert_tokens_eq(
+            &module.functions[0].to_call_rust_args(&module.swift_bridge_path, &module.types),
+            &quote! {unsafe { *Box::from_raw(arg) }},
+        );
+    }
+
     /// Verify that arguments that are owned declared types get unboxed.
     #[test]
     fn does_not_allow_owned_foreign_type_args() {
